package detail

import (
	"bufio"
	log "dbackup/logging"
	"fmt"
	"github.com/robfig/cron"
	"io"
	"io/ioutil"
	"os"
	"os/exec"
	user2 "os/user"
	"path/filepath"
	"runtime"
	"strconv"
	"strings"
	"time"
)

const (
	PGDumpCmd = "pg_dump"
)

type PGConfig struct {
	Host        string   `toml:"host"`
	Port        int      `toml:"port"`
	DB          []string `toml:"db"`
	Username    string   `toml:"username"`
	Password    string   `toml:"password"`
	Destination string   `toml:"destination"`
	Keep        int64      `toml:"keep"`
	Cron        string   `toml:"cron"`
	Options     []string `toml:"options"`
}

type Postgres struct {
	PGConfig []PGConfig
}

func (x *Postgres) Setup() {
	for _, c := range x.PGConfig {
		if strings.Index(c.Destination, "\\") != -1 {
			log.Fatal("invalid destination: %s, must use Unix Path Seperator `/`")
		}
		for _, i := range c.DB {
			if e := os.MkdirAll(c.Destination + "/" + i, os.ModePerm); e != nil {
				log.Fatal("can't mkdir: %s, err: %s", c.Destination, e)
			}
		}
	}

	user, e := user2.Current()

	if e != nil {
		log.Fatal("can't get current user: %s", e)
	}

	var pgpass string

	if runtime.GOOS == "windows" {
		pgpass = filepath.Join(os.Getenv("AppData"), "postgresql")
		_ = os.MkdirAll(pgpass, os.ModePerm)
		pgpass = filepath.Join(pgpass, "pgpass.conf")

	} else if runtime.GOOS == "linux" {

		pgpass = filepath.Join(user.HomeDir, ".pgpass")

	} else {
		log.Fatal("un-support os")
	}

	f, e := os.OpenFile(pgpass, os.O_RDWR, 0)
	if e != nil {
		f, e = os.Create(pgpass)
		if e != nil {
			log.Fatal("can't create %s, err: %s", f, e)
		}

		if runtime.GOOS == "linux" {
			_ = os.Chmod(pgpass, 0600)
		}
	}

	set := make(map[string]bool)

	scanner := bufio.NewScanner(f)
	for scanner.Scan() {
		ff := scanner.Text()
		log.Info("==>> %s", ff)
		set[ff] = true
	}

	if err := scanner.Err(); err != nil {
		f.Close()
		log.Fatal("can't read pgpass: %s", err)
	}

	for _, cfg := range x.PGConfig {
		for _, db := range cfg.DB {
			line := fmt.Sprintf("%s:%d:%s:%s:%s", cfg.Host, cfg.Port, db, cfg.Username, cfg.Password)
			set[line] = true
		}
	}

	e = f.Truncate(0)
	f.Seek(0, io.SeekStart)
	for k, _ := range set {
		_, _ = fmt.Fprintf(f, "%s\n", k)
	}

	f.Close()

	// NOTE: don't work when run as service
	// hard code, fuck windows
	// assume that %APPDATA% is C:\Users\Administrator\AppData\Roaming
	//if runtime.GOOS == "windows" {
	//	r := "C:/Users/Administrator/AppData/Roaming/postgresql"
	//	_ = os.MkdirAll(r, os.ModePerm)
	//	e = os.Rename(pgpass, r + "/pgpass.conf")
	//	if e != nil {
	//		log.Fatal("can't rename: %s", e)
	//	}
	//}
}

func (x *Postgres) AddTask(t *Task) error {
	x.Setup()
	log.Info("postgres adding task, count: %d", len(x.PGConfig))
	for _, cfg := range x.PGConfig {
		if e := cfg.AddCron(t.Cron); e != nil {
			return e
		}
	}
	return nil
}

func (x *PGConfig) Run() {
	log.Info("backup postgresql...")
	for _, db := range x.DB {
		path := fmt.Sprintf(`%s/%s/%v_%v.dump`, x.Destination, db, db, time.Now().Format("20060102-150405"))
		options := append(x.GetOptions(db), "-f", path) //fmt.Sprintf(`-f%v`, path))
		cmd := fmt.Sprintf("%s %s", PGDumpCmd, strings.Join(options, " "))
		log.Info("running task: `%s`", cmd)
		out, err := exec.Command(PGDumpCmd, options...).CombinedOutput()
		if err != nil {
			log.Fatal("failed: %s, %s", err, out)
		} else {
			log.Info("ok")
			x.CleanUp(db)
		}
	}
}

func (x *PGConfig) CleanUp(db string) {
	root := fmt.Sprintf("%s/%s", x.Destination, db)
	info, e := ioutil.ReadDir(root)
	if e != nil {
		log.Fatal("can't list dir: %s, err: %s", x.Destination, e)
	}

	for _, i := range info {
		if strings.Index(i.Name(), db) == 0 {
			f := fmt.Sprintf("%s/%s", root, i.Name())
			o, e := os.Stat(f)
			if e != nil {
				log.Error("can't get last modify time for: %s, err: %s", f, e)
				continue
			}
			if time.Since(o.ModTime()) >= time.Duration(x.Keep * int64(time.Hour)) {
				log.Info("remove expired backup: %s", i.Name())
				_ = os.Remove(f)
			}
		}
	}
}

func (x *PGConfig) AddCron(c *cron.Cron) error {
	s := strings.Split(x.Cron, " ")
	i := 0
	for j := 0; j < len(s); j++ {
		if len(strings.TrimSpace(s[j])) != 0 {
			s[i] = s[j]
			i += 1
		}
	}
	s = s[0:i]
	log.Info("%s", s)
	if len(s) != 6 {
		return fmt.Errorf("invalid cron format: %s", x.Cron)
	}
	return c.AddJob(strings.Join(s, " "), x)
}

func (x *PGConfig) GetOptions(db string) []string {
	options := x.Options

	options = append(options, "-d", db, "-h", x.Host, "-p", strconv.Itoa(x.Port), "-U", x.Username)

	return options
}